import Combine
import Foundation

public enum Bus {
  private static let e = PassthroughSubject<(key: String, value: Any), Never>()

  public static var events: AnyPublisher<(key: String, value: Any), Never> {
    e.eraseToAnyPublisher()
  }
}

public extension Bus {
  /// Асинхронно обрабатывать входящие события из шины.
  static func receiveAsync<T>(
    _ keys: Set<String>,
    _ handler: @escaping ((String, T) -> Void),
    _ subscriptions: inout [AnyCancellable]
  ) {
    Self.events
      .compactMap { convertKeyValue(keys, $0) }
      .receive(on: DispatchQueue.main)
      .sink { v in handler(v.0, v.1) }
      .store(in: &subscriptions)
  }
    
  /// Синхронно обрабатывать входящие события из шины.
  static func receiveSync<T>(
    _ keys: Set<String>,
    _ handler: @escaping ((String, T) -> Void),
    _ subscriptions: inout [AnyCancellable]
  ) {
    Self.events
      .compactMap { convertKeyValue(keys, $0) }
      .sink { v in handler(v.0, v.1) }
      .store(in: &subscriptions)
  }

  /// Асинхронно отправлять события из узла в шину.
  static func sendAsync<T>(
    _ key: String,
    _ node: AnyPublisher<T, Never>,
    _ subscriptions: inout [AnyCancellable]
  ) {
    node
      .receive(on: DispatchQueue.main)
      .sink { v in Self.e.send((key, v)) }
      .store(in: &subscriptions)
  }
    
  /// Синхронно отправлять события из узла в шину.
  static func sendSync<T>(
    _ key: String,
    _ node: AnyPublisher<T, Never>,
    _ subscriptions: inout [AnyCancellable]
  ) {
    node
      .sink { v in Self.e.send((key, v)) }
      .store(in: &subscriptions)
  }
  
  /// Синхронно отправить событие в шину один раз.
  static func send(_ key: String, _ value: Any) {
    Self.e.send((key, value))
  }
}

public extension Bus {
  static func processAsync<Src, Dst>(
    _ keysIn: Set<String>,
    _ keyOut: String,
    _ handler: @escaping ((Src) -> Dst?),
    _ subscriptions: inout [AnyCancellable]
  ) {
    Self.events
    .compactMap { processKeysValue($0, keysIn, handler) }
    .receive(on: DispatchQueue.main)
    .sink { vOut in Self.e.send((keyOut, vOut)) }
    .store(in: &subscriptions)
  }

  static func processSync<Src, Dst>(
    _ keysIn: Set<String>,
    _ keyOut: String,
    _ handler: @escaping ((Src) -> Dst?),
    _ subscriptions: inout [AnyCancellable]
  ) {
    Self.events
    .compactMap { processKeysValue($0, keysIn, handler) }
    .sink { vOut in Self.e.send((keyOut, vOut)) }
    .store(in: &subscriptions)
  }
}
